#!/usr/bin/env python3.7

# The functions in this script is used to manage submitting jobs to different clusters
# This script is called by main.py
import pandas as pd
import os
import re
import subprocess
import time

# other modules
from TileSeqMut import alignment

# phix reference sequence
phix = ""


def alignment_sh_guru(fastq_map, ref_name, ref_seq, ref_path, sam_path, ds_sam_path, sh_output):
    """
    fastq_map: df contains paths to fastq files and downsamled fastq files
    ref_name: name for the reference sequence (same as project name)
    ref_seq: reference sequence to put in fasta file
    ref_path: build fasta in this path
    sam_path: path to sam files
    ds_sam_path: path to downsampled sam files
    sh_output: make sh files to submit to the cluster

    make reference for bowtie2
    make SGE sh files
    submit sh files to SGE
    """
    ## build reference
    ref = alignment.make_ref(ref_name, ref_seq, ref_path)

    for index, row in fastq_map.iterrows():
        sample_name = os.path.basename(row["R1"]).split("_")[0]

        shfile = os.path.join(sh_output, sample_name+"_aln.sh") # for each sample, the alignment is for both R1 and R2 (they are aligning separately)
        log_file = alignment.align_main(ref, row["R1"], row["R2"], sam_path, shfile)

        sub_cmd = f"qsub -cwd -N {'aln_'+sample_name} -e {log_file} {shfile}"
        os.system(sub_cmd)
        shfile_ds = os.path.join(sh_output, sample_name+"_aln_ds.sh")
        log_file_ds = alignment.align_main(ref, row["r1_ds"], row["r2_ds"], ds_sam_path, shfile_ds)
        sub_cmd = f"qsub -cwd -N {'aln_ds_'+sample_name} -e {log_file_ds} {shfile_ds}"
        os.system(sub_cmd)
        #break

def alignment_sh_galen(fastq_map, ref_name, ref_seq, ref_path, sam_path, sh_output, at, logging, rc):
    """
    fastq_map: df contains paths to fastq files and downsamled fastq files
    ref_name: name for the reference sequence (same as project name)
    ref_seq: reference sequence to put in fasta file
    ref_path: build fasta in this path
    sam_path: path to sam files
    ds_sam_path: path to downsampled sam files
    sh_output: make sh files to submit to the cluster

    make reference for bowtie2
    make SGE sh files
    submit sh files to SGE
    """
    ## build reference
    ref = alignment.make_ref(ref_name, ref_seq, ref_path)
    phix = alignment.make_ref("phix", ref_seq, ref_path)

    # store sam paths
    fastq_map = pd.concat([fastq_map, pd.DataFrame(columns=["r1_sam", "r2_sam"])])
    all_job_id = []
    for index, row in fastq_map.iterrows():
        sample_name = os.path.basename(row["R1"]).split("_")[0]

        # for each sample, the alignment is for both R1 and R2 (they are aligned separately)
        shfile = os.path.join(sh_output, sample_name+"_aln.sh")

        # write header to sh file
        # assume at is single digit
        time_request = f"0{at}:00:00"
        # create log file for alignment
        sam_log_f = os.path.join(sam_path, f"{sample_name}")
        header = f"#!/bin/bash\n#SBATCH --time={time_request}\n#SBATCH --job-name={sample_name}\n#SBATCH " \
                 f"--error={sam_log_f}-%j.log\n#SBATCH --output={sam_log_f}-%j.log\n"

        if "Undetermined" in sample_name:
            r1_sam, r2_sam, log_file = alignment.align_main(phix, row["R1"], row["R2"], sam_path, shfile, rc=rc,
                                                            header=header)
        else:
            r1_sam, r2_sam, log_file = alignment.align_main(ref, row["R1"], row["R2"], sam_path, shfile, rc=rc, header=header)

        row["r1_sam"] = r1_sam
        row["r2_sam"] = r2_sam
        sub_cmd = ["sbatch", str(shfile)]
        jobs = subprocess.run(sub_cmd, stdout=subprocess.PIPE)
        job_id = jobs.stdout.decode("utf-8").strip().split(".")[0]
        all_job_id.append(job_id.split()[-1])
        logging.info(f"{sample_name}: job id - {job_id}")
    return fastq_map, all_job_id


def alignment_sh_bc2(fastq_map, ref_name, ref_seq, ref_path, sam_path, sh_output, at, logging, rc):
    """
    submit jobs to BC/BC2
    return a df with columns: [R1, R2, r1_sam, r2_sam]
    return a list of job id that we just submited
    """

    # build reference
    ref = alignment.make_ref(ref_name, ref_seq, ref_path)
    # store sam paths
    fastq_map = pd.concat([fastq_map, pd.DataFrame(columns=["r1_sam", "r2_sam"])])
    time = at
    all_job_id = []
    for index, row in fastq_map.iterrows(): # go through all the fastq pairs
        sample_name = os.path.basename(row["R1"]).split("_")[0]

        shfile = os.path.join(sh_output, f"Aln_{sample_name}.sh")
        r1_sam, r2_sam, log_file = alignment.align_main(ref, row["R1"], row["R2"], sam_path, shfile, rc=rc)
        row["r1_sam"] = r1_sam
        row["r2_sam"] = r2_sam
        # create log file for alignment
        sam_log_f = os.path.join(sam_path, f"{sample_name}.log")
        sub_cmd = ["submitjob2", "-w", str(time), "-c", "1", str(shfile), "2>", sam_log_f]
        jobs = subprocess.run(sub_cmd, stdout=subprocess.PIPE)

        job_id = jobs.stdout.decode("utf-8").strip().split(".")[0]
        all_job_id.append(job_id)
        logging.info(f"{sample_name}: job id - {job_id}")

    return fastq_map, all_job_id


def alignment_sh_dc(fastq_map, ref_name, ref_seq, ref_path, sam_path, sh_output, at, logging, rc):
    """
    submit jobs to BC/BC2
    return a df with columns: [R1, R2, r1_sam, r2_sam]
    return a list of job id that we just submited
    """

    # build reference
    ref = alignment.make_ref(ref_name, ref_seq, ref_path)
    # store sam paths
    fastq_map = pd.concat([fastq_map, pd.DataFrame(columns=["r1_sam", "r2_sam"])])
    time = at
    all_job_id = []
    for index, row in fastq_map.iterrows(): # go through all the fastq pairs
        sample_name = os.path.basename(row["R1"]).split("_")[0]

        shfile = os.path.join(sh_output, f"Aln_{sample_name}.sh")
        r1_sam, r2_sam, log_file = alignment.align_main(ref, row["R1"], row["R2"], sam_path, shfile, rc=rc)
        row["r1_sam"] = r1_sam
        row["r2_sam"] = r2_sam
        # create log file for alignment
        sam_log_f = os.path.join(sam_path, f"{sample_name}.log")
        sub_cmd = ["submitjob","-w", str(time),"-c", "1", str(shfile), "2>", sam_log_f]
        jobs = subprocess.run(sub_cmd, stdout=subprocess.PIPE)
        job_id = jobs.stdout.decode("utf-8").strip().split(".")[0]
        all_job_id.append(job_id)
        logging.info(f"Sample {sample_name}: job id - {job_id}")

    return fastq_map, all_job_id


def mut_count_sh_bc(sample_name, cmd, mt, mm, sh_output_dir, logger, cores):
    """
    Submit mutation count jobs to BC

    """
    # go through files df and submit jobs for each pair of sam files
    # counting mutations in raw sam output files
    shfile = os.path.join(sh_output_dir, f"Mut_count_{sample_name}.sh")
    log_f = os.path.join(sh_output_dir, f"Mut_count_{sample_name}.log")
    with open(shfile, "w") as sh:
        sh.write(cmd+"\n")
        os.system(f"chmod 755 {shfile}")
    # submit this to the cluster
    sub_cmd = ["submitjob2","-w", str(mt), "-c", f"{cores}", "-m", f"{mm}", shfile, "&>>", log_f]
    logger.debug(sub_cmd)
    job = subprocess.run(sub_cmd, stdout=subprocess.PIPE)
    job_id = job.stdout.decode("utf-8").strip().split(".")[0]
    # log sample name and job id
    logger.info(f"Sample {sample_name}: job id - {job_id}")
    return job_id


def mut_count_sh_dc(sample_name, cmd, mt, mm, sh_output_dir, logger, cores):
    """
    Submit mutation count jobs to DC
    """
    # go through files df and submit jobs for each pair of sam files
    # counting mutations in raw sam output files
    shfile = os.path.join(sh_output_dir, f"Mut_count_{sample_name}.sh")
    log_f = os.path.join(sh_output_dir, f"Mut_count_{sample_name}.log")
    with open(shfile, "w") as sh:
        sh.write(cmd+"\n")
        os.system(f"chmod 755 {shfile}")
    #sample_error_file = os.path.join(log_dir, f"sample_{sample_name}.log")
    # submit this to the cluster
    sub_cmd = ["submitjob", "-w", str(mt), "-c", f"{cores}", "-m", f"{mm}", shfile, "&>>", log_f]
    logger.debug(sub_cmd)
    job = subprocess.run(sub_cmd, stdout=subprocess.PIPE)
    job_id = job.stdout.decode("utf-8").strip()
    # log sample name and job id
    logger.info(f"Sample {sample_name}: job id - {job_id}")
    return job_id


def mut_count_sh_galen(sample_name, cmd, mt, mm, sh_output_dir, logger, cores):
    """
    Submit mutation count jobs to DC
    """
    # go through files df and submit jobs for each pair of sam files
    # counting mutations in raw sam output files
    shfile = os.path.join(sh_output_dir, f"Mut_count_{sample_name}.sh")
    log_f = os.path.join(sh_output_dir, f"Mut_count_{sample_name}")
    time_request = f"{mt}:00:00"

    header = f"#!/bin/bash\n#SBATCH --time={time_request}\n#SBATCH --job-name={sample_name}\n#SBATCH " \
             f"--cpus-per-task={cores}\n#SBATCH --error={log_f}-%j.log\n#SBATCH --mem={mm}G\n#SBATCH " \
             f"--output={log_f}-%j.log\n"

    with open(shfile, "w") as sh:
        sh.write(header)
        sh.write(cmd+"\n")
        os.system(f"chmod 755 {shfile}")
    #sample_error_file = os.path.join(log_dir, f"sample_{sample_name}.log")
    # submit this to the cluster
    sub_cmd = ["sbatch", str(shfile)]
    logger.debug(sub_cmd)
    job = subprocess.run(sub_cmd, stdout=subprocess.PIPE)
    job_id = job.stdout.decode("utf-8").strip()
    # log sample name and job id
    logger.info(f"Sample {sample_name}: job id - {job_id}")
    return job_id.split()[-1]


def parse_jobs(job_list, env, logger):
    """
    return true if all the jobs in job list finished
    else wait for 10 mins and return how man jobs are running and queued
    job_list: list of job ids
    logger: logging object
    """
    qstat_cmd = ["qstat"] + job_list
    job = subprocess.run(qstat_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    qstat_out = job.stdout.decode("utf-8", errors="replace")
    qstat_err = job.stderr.decode("utf-8", errors="replace")

    f_id = []
    updated_list = []
    while True:
        running = []
        queued = []
        completed = []

        if qstat_err != "":
            # jobs might be finished and no longer in queue
            # extract job id from std err
            if env == "BC" or env == "BC2":
                err = qstat_err.split("\n")[:-1]
                id_regex = re.compile(r"(\d+).bc")
            elif env == "DC":
                err = qstat_err.split("\n")[:-1]
                id_regex = re.compile(r"(\d+).dc[0-9]+")

            f_id = [] # finished jobs
            for i in err:
                try:
                    match = id_regex.search(i)
                    job_id = match.group(1)
                    f_id.append(job_id)
                except:
                    logger.warning(i)
                    continue
            err_id = set(f_id)
            updated_list = [x for x in job_list if x not in err_id]

        if qstat_out != "":
            qstat_out = qstat_out.split("\n")[:-1]
            if env == "BC" or env == "BC2":
                id_regex = re.compile(r"(\d+).bc.+(R|Q|C|E)")
            elif env == "DC":
                id_regex = re.compile(r"(\d+).dc[0-9]+.+(R|Q|C|E)")

            for line in qstat_out:
                if ("---" in line) or ("Job ID" in line): continue
                match = id_regex.search(line)
                job_id = match.group(1)
                job_s = match.group(2)
                if job_s == "E" or job_s == "C":
                    completed.append(job_id)
                elif job_s == "R":
                    running.append(job_id)
                elif job_s == "Q":
                    queued.append(job_id)

        logger.info(f"{len(queued)} jobs queued")
        logger.info(f"{len(running)} jobs running")
        final_list = list(set(updated_list+running+queued))
        if final_list == []:
            return True
        else:
            # check in 10min
            time.sleep(600)
            job_list = final_list
            qstat_cmd = ["qstat"] + job_list
            job = subprocess.run(qstat_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            qstat_out = job.stdout.decode("utf-8", errors="replace")
            qstat_err = job.stderr.decode("utf-8", errors="replace")


def parse_jobs_galen(job_list, logger):
    """
    Galen uses slurm scheduler, different from BC and DC
    return true if all the jobs in job list finished
    else wait for 10 mins and return how man jobs are running and queued
    job_list: list of job ids
    logger: logging object
    """
    qstat_cmd = ["squeue", "-j", ",".join(job_list)]
    job = subprocess.run(qstat_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    qstat_out = job.stdout.decode("utf-8", errors="replace")
    qstat_err = job.stderr.decode("utf-8", errors="replace")
    logger.debug(qstat_out)
    logger.debug(qstat_err)
    while True:
        running_jobs, queued_jobs, completed_jobs = [], [], []
        if qstat_out != "":
            # make df
            qstat_df = pd.DataFrame([i.split() for i in qstat_out.split("\n")])
            qstat_df = qstat_df.rename(columns=qstat_df.iloc[0])
            qstat_df = qstat_df.drop(qstat_df.index[1])
            logger.debug(qstat_df)
            # get all active job ID
            running_jobs = qstat_df[qstat_df["ST"] == "R"]["JOBID"].tolist()
            # in queue
            queued_jobs = qstat_df[qstat_df["ST"] == "PD"]["JOBID"].tolist()
            # completing
            completed_jobs = qstat_df[qstat_df["ST"] == "CG"]["JOBID"].tolist()
            logger.debug(running_jobs)
            logger.debug(queued_jobs)

        logger.info(f"{len(queued_jobs)} jobs queued")
        logger.info(f"{len(running_jobs)} jobs running")

        final_list = list(set(running_jobs + queued_jobs))
        if final_list == []:
            return True
        else:
            # check in 10min
            time.sleep(600)
            job_list = final_list
            qstat_cmd = ["squeue", "-j", ",".join(job_list)]
            job = subprocess.run(qstat_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            qstat_out = job.stdout.decode("utf-8", errors="replace")
            qstat_err = job.stderr.decode("utf-8", errors="replace")


def submit_given_jobs(shfile, logger, mt, mm, cores, env=""):
    """
    for a given sh file, submit to cluster, return job id
    """
    sample_name = shfile.split(".")[0].split("_")[-1]
    sh_dir = os.path.dirname(shfile)
    log_f = os.path.join(sh_dir, f"Mut_count_{sample_name}.log")
    if env == "GALEN":
        sub_cmd = ["sbatch", str(shfile)]
        job = subprocess.run(sub_cmd, stdout=subprocess.PIPE)
        job_id = job.stdout.decode("utf-8").strip().split()[-1]
    elif env == "BC2" or env == "BC":
        sub_cmd = ["submitjob2", "-w", str(mt), "-c", f"{cores}", "-m", f"{mm}", shfile, "&>>", log_f]
        job = subprocess.run(sub_cmd, stdout=subprocess.PIPE)
        job_id = job.stdout.decode("utf-8").strip()
    elif env == "DC":
        sub_cmd = ["submitjob", "-w", str(mt), "-c", f"{cores}", "-m", f"{mm}", shfile, "&>>", log_f]
        job = subprocess.run(sub_cmd, stdout=subprocess.PIPE)
        job_id = job.stdout.decode("utf-8").strip()
    else:
        raise ValueError("Wrong environment code")
    # log sample name and job id
    logger.info(f"Sample {sample_name}: job id - {job_id}")
    return job_id


if __name__ == "__main__":
    # test job list
    job_list = ["42091", "42090", "42089", "42080"]
    # parse_jobs(job_list, "DC", "")
    parse_jobs_galen(job_list, "")



#!/usr/bin/env python3.7

## Read sam file (R1 and R2)
# count mutations in sam files
# output mutation counts

import pandas as pd
import re
import os
import math
import logging
import mmap
import time
import glob
import multiprocessing as mp
import argparse
# import datetime
from collections import deque

# modules in package
from TileSeqMut import help_functions
from TileSeqMut import locate_mut

# import help_functions
# import locate_mut

class readSam(object):

    def __init__(self, sam_r1, sam_r2, param, arguments, output_dir, cores, loggerobj):
        """
        sam_R1: read one of the sample
        sam_R2: read two of the sample

        output_dir: main output directory

        log_level: settings for logging
        log_dir: directory to save all the logging for each sample
        """
        self._r1 = sam_r1
        self._r2 = sam_r2
        self._param = param
        self._project, self._seq, self._cds_seq, self._tile_map, self._region_map, self._samples, self._var, self._relations = help_functions.parse_json(param)

        self._qual = self._var["posteriorThreshold"]
        min_cover = self._var["minCover"]
        self._mutrate = self._var["mutRate"]
        self._output_counts_dir = output_dir
        self._cores = cores

        # sample information
        self._sample_id = os.path.basename(sam_r1).split("_")[0]
        self._sample_info = self._samples[self._samples["Sample ID"] == self._sample_id]

        # tile information
        self._sample_tile = self._sample_info["Tile ID"].values[0]
        self._tile_begins = (self._tile_map[self._tile_map["Tile Number"] == self._sample_tile]["Start AA"].values[0] *3)-2 # beginning position of this tile (cds position)
        self._tile_ends = self._tile_map[self._tile_map["Tile Number"] == self._sample_tile]["End AA"].values[0] *3  # ending position of this tile (cds position)
        self._tile_len = self._tile_ends - self._tile_begins
        self._cds_start = self._seq.cds_start
        self._min_map_len = math.ceil(self._tile_len * float(min_cover))

        self._sample_condition = self._sample_info["Condition"].values[0]
        self._sample_tp = self._sample_info["Time point"].values[0]
        self._sample_rep = int(self._sample_info["Replicate"].values[0])

        self._seq_lookup = help_functions.build_lookup(self._cds_start.values.item(), self._seq.cds_end.values.item(), self._cds_seq)

        # init a new object for logging, log to sample specific file.log
        # log_f = os.path.join(output_dir, f"sample_{str(self._sample_id)}_mut_count.log")
        # log_object = help_functions.logginginit(arguments.log_level, log_f)

        self._sample_counts_f = os.path.join(self._output_counts_dir,f"counts_sample_{self._sample_id}.csv")
        self._sample_counts_r1_f = os.path.join(self._output_counts_dir, f"counts_sample_{self._sample_id}_r1.csv")
        self._sample_counts_r2_f = os.path.join(self._output_counts_dir, f"counts_sample_{self._sample_id}_r2.csv")

        self._base = arguments.base
        self._posteriorQC = arguments.posteriorQC
        # self._mut_log = log_object.getLogger("count.mut")
        # self._locate_log = log_object.getLogger("locate.mut")
        self._mut_log = loggerobj
        self._mut_log.info(f"Counting mutations in sample-{self._sample_id}")
        self._mut_log.info(f"Sam file input R1:{sam_r1}")
        self._mut_log.info(f"Sam file input R2:{sam_r2}")

        output_csv = open(self._sample_counts_f, "w")
        # write log information to counts output
        output_csv.write(f"#Sample:{self._sample_id}\n#Tile:{self._sample_tile}\n#Tile Starts:{self._tile_begins}\n"
                         f"#Tile Ends:{self._tile_ends}\n#Condition:{self._sample_condition}\n"
                         f"#Replicate:{self._sample_rep}\n#Timepoint:{self._sample_tp}\n"
                         f"#Posterior cutoff:{self._qual}\n#min cover %:{min_cover}\n")
        output_csv.close()

        if self._posteriorQC:
            output_csv = open(self._sample_counts_r1_f, "w")
            # write log information to counts output
            output_csv.write(f"#Sample:{self._sample_id}\n#Tile:{self._sample_tile}\n#Tile Starts:{self._tile_begins}\n"
                             f"#Tile Ends:{self._tile_ends}\n#Condition:{self._sample_condition}\n"
                             f"#Replicate:{self._sample_rep}\n#Timepoint:{self._sample_tp}\n"
                             f"#Posterior cutoff:{self._qual}\n#min cover %:{min_cover}\n")
            output_csv.close()

            output_csv = open(self._sample_counts_r2_f, "w")
            # write log information to counts output
            output_csv.write(f"#Sample:{self._sample_id}\n#Tile:{self._sample_tile}\n#Tile Starts:{self._tile_begins}\n"
                             f"#Tile Ends:{self._tile_ends}\n#Condition:{self._sample_condition}\n"
                             f"#Replicate:{self._sample_rep}\n#Timepoint:{self._sample_tp}\n"
                             f"#Posterior cutoff:{self._qual}\n#min cover %:{min_cover}\n")
            output_csv.close()

        # build a df for tracking how many reads passed filter at certain position of the read 
        # df contains ["pos", "m_r1", "m_r2", "passed"] columns
        # positions referes to all the nt positions in this tile 
        self._track_reads = pd.DataFrame({}, columns=["pos", "m_r1", "m_r2", "m_both", "passed"])
        
        self._track_reads["pos"] = range(self._tile_begins, self._tile_ends+1)
        self._track_reads = self._track_reads.set_index("pos")
        self._track_reads = self._track_reads.fillna(0)

    def adjust_er(self, wt_override=False):
        """
        Based on the wt samples given in the parameter sheet,
        calculate new error rate for phred scores
        this function is dependent on the tileseqMave package
        please make sure the correct script is installed before running the pipeline
        """
        r_df = pd.DataFrame(self._relations)
        phred_output = []
        
        if r_df.empty:
            if not wt_override:
                # no wt relationship is defined
                self._mut_log.warning("No WT relationship! No error rate correction can be applied")
                return phred_output

            else:  # treating EVERYTHING in the run as wt
                self._mut_log.warning("WT OVERRIDE ON, TREATING EVERYTHING AS WT")
                wt = pd.DataFrame({}, columns=["Condition 1", "Relationship", "Condition 2"])
                wt["Condition 1"] = self._samples["Condition"]
        else:
            # find out wt sample ID for this sample
            wt = r_df[r_df["Relationship"]=="is_wt_control_for"]
            if wt.empty:
                if not wt_override:
                    # no wt relationship is defined
                    self._mut_log.warning("No WT relationship! No error rate correction can be applied")
                    return phred_output

                else:  # treating EVERYTHING in the run as wt
                    self._mut_log.warning("WT OVERRIDE ON, TREATING EVERYTHING AS WT")
                    wt["Condition 1"] = self._samples["Condtion"]
        wt_id = ""

        if self._sample_condition in wt["Condition 1"].tolist():
            if self._sample_rep == 1:
                # this sample is wt 
                # run calibratePhred 
                # we only need 1 wt calibration for each tile (r1 and r2)
                # we are always using the first rep 
                wt_id = self._sample_id
            else:
                # this is rep 2 of the wt
                # find corresponding rep 1 of the wt 
                wt_id = self._samples[(self._samples["Condition"] == self._sample_condition) & (self._samples["Tile ID"] == self._sample_tile) & (self._samples["Replicate"] == 1)]["Sample ID"].values[0]

        elif self._sample_condition in wt["Condition 2"].tolist():
            # get corresponding wt 
            wt_name = wt[wt["Condition 2"] == self._sample_condition]["Condition 1"].values[0]
            # get wt sample for this tile and rep 1
            wt_id = self._samples[(self._samples["Condition"] == wt_name) & (self._samples["Tile ID"] == self._sample_tile) & (self._samples["Replicate"] == 1)]["Sample ID"].values[0]
        
        # check if calibrated
        self._mut_log.info(f"wt sample used to adjust phred scores: {wt_id}")
        phred_output_r1 = os.path.join(self._output_counts_dir, f"{wt_id}_T{self._sample_tile}_R1_calibrate_phred.csv")
        phred_output_r2 = os.path.join(self._output_counts_dir, f"{wt_id}_T{self._sample_tile}_R2_calibrate_phred.csv")
        if not os.path.isfile(phred_output_r1):
            # create an empty file as place holder 
            with open(phred_output_r1, 'w') as fp:
                pass
            log_f = os.path.join(self._output_counts_dir, f"{wt_id}_R1_phred.log")
            cmd_r1 = f"calibratePhred.R {self._r1} -p {self._param} -o {phred_output_r1} -l {log_f} --silent --cores {self._cores}"
            os.system(cmd_r1)
        if not os.path.isfile(phred_output_r2):
            # create an empty file as place holder 
            with open(phred_output_r2, 'w') as fp:
                pass
            log_f = os.path.join(self._output_counts_dir, f"{wt_id}_R2_phred.log")
            cmd_r2 = f"calibratePhred.R {self._r2} -p {self._param} -o {phred_output_r2} -l {log_f} --silent --cores {self._cores}"
            os.system(cmd_r2)
        
        # check if both file has something in there
        # if they are empty, wait for them to finish (they might be running in other jobs
        t0 = time.time()  # check for timeout
        while os.stat(phred_output_r1).st_size == 0:
            os.system("sleep 300")
            self._mut_log.warning("phred file (R1) found, but empty.. Waiting for the job to finish...")
            t1 = time.time()
            total = float(t1-t0)
            if total > 10800:
                raise TimeoutError(f"Wating for Phred file {phred_output_r1} timeout")
        t0 = time.time()
        while os.stat(phred_output_r2).st_size == 0:
            os.system("sleep 300")
            self._mut_log.warning("phred file (R2) found, but empty.. Waiting for the job to finish...")
            t1 = time.time()
            total = float(t1-t0)
            if total > 10800:
                raise TimeoutError(f"Wating for Phred file {phred_output_r2} timeout")
        time.sleep(30)
        self._mut_log.info(f"Adjusted thred files generated: {phred_output_r1}, {phred_output_r2}")
        return [phred_output_r1, phred_output_r2]

    def adjust_er_phix(self):
        """
        If not phix thred adjusted, run calibratePhred.R with phix reads
        Before running calibrate phred, shrink sam file size by selecting reads aligned to phix only
        """
        dir_path = os.path.dirname(os.path.realpath(__file__))
        phix_fasta = os.path.join(dir_path, "data/phix.fasta")
        tmp_header = os.path.join(self._output_counts_dir, "fakeheader")
        sam_dir = os.path.dirname(self._r1)
        phix_r1 = glob.glob(f"{sam_dir}/Undetermined_*_R1_*.sam")[0]
        phix_r2 = glob.glob(f"{sam_dir}/Undetermined_*_R2_*.sam")[0]
        with open(tmp_header, "w") as sam_header:
            sam_header.write("@SQ	SN:phiX	LN:5386")

        phred_output_r1 = os.path.join(self._output_counts_dir, f"phix_R1_calibrate_phred.csv")
        phred_output_r2 = os.path.join(self._output_counts_dir, f"phix_R2_calibrate_phred.csv")
        if not os.path.isfile(phred_output_r1):

            # create an empty file as place holder
            with open(phred_output_r1, 'w') as fp:
                pass
            # trim the sam file
            trimmed = phix_r1.replace(".sam", "_trimmed.sam")
            cmd = f"cat {tmp_header} {phix_r1} | samtools view -S -F 4 - -o {trimmed}"
            os.system(cmd)

            log_f = os.path.join(self._output_counts_dir, f"phix_phred.log")
            cmd = f"calibratePhred.R {phix_r1} -p {self._param} -o {phred_output_r1} -l {log_f} --silent --cores {self._cores} --fastaref {phix_fasta}"
            os.system(cmd)
        if not os.path.isfile(phred_output_r2):
            # create an empty file as place holder
            with open(phred_output_r2, 'w') as fp:
                pass
            # trim the sam file
            trimmed = phix_r2.replace(".sam", "_trimmed.sam")
            cmd = f"cat {tmp_header} {phix_r2} | samtools view -S -F 4 - -o {trimmed}"
            os.system(cmd)

            log_f = os.path.join(self._output_counts_dir, f"phix_phred.log")
            cmd = f"calibratePhred.R {phix_r2} -p {self._param} -o {phred_output_r2} -l {log_f} --silent --cores" \
                  f" {self._cores} --fastaref {phix_fasta}"
            os.system(cmd)

        # check if both file has something in there
        # if they are empty, wait for them to finish (they might be running in other jobs
        t0 = time.time()  # check for timeout
        while os.stat(phred_output_r1).st_size == 0:
            os.system("sleep 300")
            self._mut_log.warning("phred file (R1) found, but empty.. Waiting for the job to finish...")
            t1 = time.time()
            total = float(t1 - t0)
            if total > 10800:
                raise TimeoutError(f"Wating for Phred file {phred_output_r1} timeout")
        t0 = time.time()
        while os.stat(phred_output_r2).st_size == 0:
            os.system("sleep 300")
            self._mut_log.warning("phred file (R2) found, but empty.. Waiting for the job to finish...")
            t1 = time.time()
            total = float(t1 - t0)
            if total > 10800:
                raise TimeoutError(f"Wating for Phred file {phred_output_r2} timeout")
        time.sleep(30)
        self._mut_log.info(f"Adjusted thred files generated: {phred_output_r1}, {phred_output_r2}")
        return [phred_output_r1, phred_output_r2]


    def multi_core(self, adjusted_er):
        """
        Read two sam files at the same time, store mutations that passed filter
        """
        read_pair = 0  # total pairs
        un_map = 0  # total number of unmapped reads
        read_nomut = 0  # read pairs that have no mutations

        hgvs_output = {}  # save all the hgvs in this pair of sam files
        r1_pop_hgvs = {}  # save all the hgvs that are 2/3nt changes based only on R1
        r2_pop_hgvs = {}  # save all the hgvs that are 2/3nt changes based only on R2
        off_mut = {}  # save all the positions that were mapped but outside of the tile
        row = {}  # create a dictionary to save all the reads information to pass on to locate_mut.py

        final_pairs = 0
        r1_popmut = 0
        r2_popmut = 0
        off_read = 0

        r1_f = open(self._r1, "r")
        r2_f = open(self._r2, "r")
        # init objects
        pool = mp.Pool(self._cores)
        jobs = []
        self._mut_log.info("Start reading SAM files")
        for line_r1, line_r2 in zip(r1_f, r2_f):

            line_r1 = line_r1.split()
            line_r2 = line_r2.split()
            if len(line_r1) < 11 or len(line_r2) < 11:
                # the read has no sequence
                self._mut_log.warning(line_r1)
                self._mut_log.warning(line_r2)
                self._mut_log.warning("Missing fields in read!")
                continue

            read_pair += 1  # count how many read pairs in this pair of sam files
            mapped_name_r1 = line_r1[2]
            mapped_name_r2 = line_r2[2]
            if mapped_name_r1 == "*" or mapped_name_r2 == "*":  # if one of the read didn't map to ref
                un_map += 1
                continue

            # check if read ID mapped
            read_name_r1 = line_r1[0]
            read_name_r2 = line_r2[0]
            if read_name_r1 != read_name_r2:
                print(read_name_r1)
                print(read_name_r2)
                self._mut_log.error("Read pair IDs did not map, please check fastq files")
                exit(1)

            # get starting position for r1 and r2
            pos_start_r1 = line_r1[3]
            pos_start_r2 = line_r2[3]
            # record reads that are not mapped to this tile
            # this is defined as if the read covers at least min_map percent of the tile
            # the default min_map is 70%
            # we also assume that the read len is ALWAYS 150
            r1_end = int(pos_start_r1) + 150
            # r1_end must cover from start of the tile to 70% of the tile
            if ((r1_end - int(self._cds_start)) < (int(self._tile_begins) + int(self._min_map_len))) or \
                    ((int(pos_start_r2) - int(self._cds_start)) > (int(self._tile_ends) - int(self._min_map_len))):
                off_read += 1
                continue

            # get CIGAR string
            CIGAR_r1 = line_r1[5]
            seq_r1 = line_r1[9]
            quality_r1 = line_r1[10]

            CIGAR_r2 = line_r2[5]
            seq_r2 = line_r2[9]
            quality_r2 = line_r2[10]

            mdz_r1 = [i for i in line_r1 if "MD:Z:" in i]
            mdz_r2 = [i for i in line_r2 if "MD:Z:" in i]

            if len(mdz_r1) != 0:
                mdz_r1 = mdz_r1[0].split(":")[-1]
            else:
                mdz_r1 = ""

            if len(mdz_r2) != 0:
                mdz_r2 = mdz_r2[0].split(":")[-1]
            else:
                mdz_r2 = ""

            if ((not re.search('[a-zA-Z]', mdz_r1)) and ("I" not in CIGAR_r1) and ("D" not in CIGAR_r1)) and \
                    ((not re.search('[a-zA-Z]', mdz_r2)) and ("I" not in CIGAR_r2) and ("D" not in CIGAR_r2)):
                # if MDZ string only contains numbers
                # and no insertions shown in CIGAR string
                # means there is no mutation in this read
                # if both reads have no mutations in them, skip this pair
                read_nomut += 1
                # remove reads that have no mutations in MDZ
                continue
            # make the reads in the format of a dictionary
            # columns=["mapped_name", "pos_start", "qual", "CIGAR", "mdz","seq"])
            # row["mapped_name_r1"] = mapped_name_r1
            row["pos_start_r1"] = pos_start_r1
            row["qual_r1"] = quality_r1
            row["cigar_r1"] = CIGAR_r1
            row["mdz_r1"] = mdz_r1
            row["seq_r1"] = seq_r1

            # row["mapped_name_r2"] = mapped_name_r2
            row["pos_start_r2"] = pos_start_r2
            row["qual_r2"] = quality_r2
            row["cigar_r2"] = CIGAR_r2
            row["mdz_r2"] = mdz_r2
            row["seq_r2"] = seq_r2
            # mut_parser = locate_mut.MutParser(row, self._seq, self._cds_seq, self._seq_lookup, self._tile_begins,
                                              # self._tile_ends, self._qual, self._locate_log, self._mutrate)
            jobs.append(pool.apply_async(process_wrapper, (row, self._seq, self._cds_seq, self._seq_lookup, self._tile_begins,
                                               self._tile_ends, self._qual, self._mut_log, self._mutrate,
                                                           self._base, self._posteriorQC, adjusted_er)))
            row = {} # flush

        r1_f.close()
        r2_f.close()
        self._mut_log.info("File streamed to subprocesses, waiting for jobs to finish")
        # wait for all jobs to finish
        all_df = []

        for job in jobs:
            hgvs, outside_mut, all_posterior, hgvs_r1_clusters, hgvs_r2_clusters, track_df = job.get()
            if len(hgvs) != 0:
                final_pairs += 1
                if hgvs_output.get(hgvs, -1) != -1:
                    hgvs_output[hgvs] += 1
                else:
                    hgvs_output[hgvs] = 1

            if len(hgvs_r1_clusters) != 0:
                r1_popmut += 1
                if hgvs_r1_clusters in r1_pop_hgvs:
                    r1_pop_hgvs[hgvs_r1_clusters] += 1
                else:
                    r1_pop_hgvs[hgvs_r1_clusters] = 1

            if len(hgvs_r2_clusters) != 0:
                r2_popmut += 1
                if hgvs_r2_clusters in r2_pop_hgvs:
                    r2_pop_hgvs[hgvs_r2_clusters] += 1
                else:
                    r2_pop_hgvs[hgvs_r2_clusters] = 1

            if outside_mut != []:
                outside_mut = list(set(outside_mut))
                for i in outside_mut:
                    if not (i in off_mut):
                        off_mut[i] = 1
            if not all_posterior.empty:
                all_df.append(all_posterior)
            track_df = track_df.set_index("pos")
            track_df = track_df[track_df.index.isin(self._track_reads.index)]
            # add track df to track summary
            track_all = pd.concat([self._track_reads, track_df], axis=1).fillna(0)
            self._track_reads = track_all.groupby(by=track_all.columns, axis=1).sum()
        # clean up
        pool.close()
        self._mut_log.info("Pool closed")
        # track mutations that are not within the same tile
        # track sequencing depth for each sample
        # write this information to a tmp file
        # merge the tmp file (in main.py)

        # tmp_f = os.path.join(self._output_counts_dir, f"{self._sample_id}_tmp.csv")
        # off_mut = list(set(off_mut))
        # off_mut.sort()
        # f = open(tmp_f, "w")
        # f.write(
        #     f"sample,{self._sample_id},tile,{self._tile_begins}-{self._tile_ends},sequencing_depth,{read_pair},off_tile_reads,{off_read},off_tile_perc,{off_read / read_pair}\n")
        # f.close()

        output_csv = open(self._sample_counts_f, "a")
        output_csv.write(f"#Raw read depth:{read_pair}\n")
        output_csv.write(
            f"#Number of read pairs without mutations:{read_nomut}\n#Number of read pairs did not map to gene:{un_map}\n#Mutation positions outside of the tile:{off_mut}\n#Number of reads outside of the tile:{off_read}\n")
        output_csv.write(f"#Final read-depth:{read_pair - un_map - off_read}\n")
        output_csv.write(f"#Total read pairs with mutations:{final_pairs}\n")
        output_csv.write(
            f"#Comment: Total read pairs with mutations = Read pairs with mutations that passed the posterior threshold\n#Comment: Final read-depth = raw read depth - reads didn't map to gene - reads mapped outside of the tile\n")

        output_csv.close()

        self._mut_log.info(f"Raw sequencing depth: {read_pair}")
        self._mut_log.info(f"Number of reads without mutations:{read_nomut}")
        self._mut_log.info(f"Final read-depth: {read_pair - un_map - off_read}")
        # convert list to df with one col
        hgvs_df = pd.DataFrame.from_dict(hgvs_output, orient="index")
        hgvs_df = hgvs_df.reset_index()
        if not hgvs_df.empty:
            hgvs_df.columns = ["HGVS", "count"]
        hgvs_df.to_csv(self._sample_counts_f, mode="a", index=False)
        del hgvs_df

        # save read coverage to csv file
        cov_file = os.path.join(self._output_counts_dir, f"coverage_{self._sample_id}.csv")
        self._track_reads.to_csv(cov_file)

        if self._posteriorQC:
            r1_df = pd.DataFrame.from_dict(r1_pop_hgvs, orient="index")
            r1_df = r1_df.reset_index()
            if not r1_df.empty:
                r1_df.columns = ["HGVS", "count"]
            r1_df.to_csv(self._sample_counts_r1_f, mode="a", index=False)
            del r1_f

            r2_df = pd.DataFrame.from_dict(r2_pop_hgvs, orient="index")
            r2_df = r2_df.reset_index()
            if not r2_df.empty:
                r2_df.columns = ["HGVS", "count"]
            r2_df.to_csv(self._sample_counts_r2_f, mode="a", index=False)
            del r2_f

            self._mut_log.info(f"Reading posterior unfiltered dfs ...")
            all_f = os.path.join(self._output_counts_dir, f"{self._sample_id}_posprob_all.csv")
            for df in all_df:
                df.to_csv(all_f, mode='a', header=False, index=False)
            self._mut_log.info(f"Posterior df saved to files ... {all_f}")
            del all_df


def process_wrapper(row, seq, cds_seq, seq_lookup, tile_begins, tile_ends, qual, locate_log, mutrate, base,
                    posteriorQC, adjusted_er):
    """

    """
    mut_parser = locate_mut.MutParser(row, seq, cds_seq, seq_lookup, tile_begins, tile_ends, qual, locate_log,
                                      mutrate, base, posteriorQC, adjusted_er)
    hgvs, outside_mut, all_df, hgvs_r1_clusters, hgvs_r2_clusters, track_df = mut_parser._main()
    return hgvs, outside_mut, all_df, hgvs_r1_clusters, hgvs_r2_clusters, track_df



#!/usr/bin/env python3.7

# Author: Roujia Li
# email: Roujia.li@mail.utoronto.ca

import os
import pandas as pd
import json
import argparse
import subprocess

from TileSeqMut.help_functions import logginginit

#for two runs merge the mutation counts files into output folder

def merge_runs(input_dir1, input_dir2, map_df, output):
    """
    :param input_dir1: Directory 1 contains mutation counts files
    :param input_dir2: Directory 2 contains mutation counts files
    :param map_df: A dataframe that maps the samples to be merged
    :param output: output directory to save merged mut_counts
    """
    # make a new file

    new_samples = []
    for index, row in map_df.iterrows():
        merged_header = ""
        mut_count_f1 = os.path.join(input_dir1, f"counts_sample_{row['Sample ID_run1']}.csv")
        mut_count_f2 = os.path.join(input_dir2, f"counts_sample_{row['Sample ID_run2']}.csv")

        new_samples.append(f"{row['Sample ID_run1']}-{row['Sample ID_run2']}")
        # join headers
        with open(mut_count_f1) as file1, open(mut_count_f2) as file2:
            for line1, line2 in zip(file1, file2):
                if "#" not in line1: break
                if "Final read-depth:" in line1:
                    n_depth_1 = int(line1.split(":")[1])
                    n_depth_2 = int(line2.split(":")[1])
                    total_depth = n_depth_1+n_depth_2
                    merged_header+= f"#Final read-depth:{total_depth}\n"
                    continue
                if "Number of read pairs without mutations:" in line1:
                    n_depth_1 = int(line1.split(":")[1])
                    n_depth_2 = int(line2.split(":")[1])
                    total_depth = n_depth_1 + n_depth_2
                    merged_header += f"#Number of read pairs without mutations:{total_depth}\n"
                    continue
                if "Total read pairs with mutations:" in line1:
                    n_depth_1 = int(line1.split(":")[1])
                    n_depth_2 = int(line2.split(":")[1])
                    total_depth = n_depth_1 + n_depth_2
                    merged_header += f"#Total read pairs with mutations:{total_depth}\n"
                    continue
                merged_header += line1.strip() + "; " + line2
        output_f = open(os.path.join(output, f"counts_sample_{row['Sample ID_run1']}-{row['Sample ID_run2']}.csv"), "w")
        output_f.write(merged_header)
        # skip header and read the mutations
        df1 = pd.read_csv(mut_count_f1, skiprows=18)
        df2 = pd.read_csv(mut_count_f2, skiprows=18)
        # for each file find the corresponded wt file and read the wt
        merge = [df1, df2]
        merge = pd.concat(merge)
        merge_counts = merge.groupby(by=["HGVS"], as_index=False)["count"].sum().sort_values("count", ascending=False)
        # save header and counts to file
        merge_counts.to_csv(output_f, mode="a", index=False)
        output_f.close()
        
        cov_f1 = os.path.join(input_dir1, f"coverage_{row['Sample ID_run1']}.csv")
        cov_f2 = os.path.join(input_dir2, f"coverage_{row['Sample ID_run2']}.csv")

        # read coverage files
        # sum up two df
        cov_d1 = pd.read_csv(cov_f1).set_index("pos")
        cov_d2 = pd.read_csv(cov_f2).set_index("pos")
        df_sum = cov_d1.add(cov_d2, fill_value=0)
        output_cov = os.path.join(output, f"coverage_{row['Sample ID_run1']}-{row['Sample ID_run2']}.csv")
        df_sum.to_csv(output_cov)
    return new_samples


def read_json(json_dict):
    """
    Read json input and return data frame of samples and their conditions
    """
    samples = json_dict["samples"]
    condition_def = pd.DataFrame(json_dict["conditions"]["definitions"])
    # get nonselect conditions
    # get wt for nonselect
    try:
        select_relation = condition_def.loc[condition_def["Relationship"].str.contains("is_selection_for")]
        wt_relation = condition_def.loc[condition_def["Relationship"].str.contains("is_wt_control_for")]
        select_relation = select_relation[["Condition 1", "Condition 2"]]
        wt_relation = wt_relation[["Condition 1", "Condition 2"]]
        relation = pd.merge(select_relation, wt_relation, how="left", left_on="Condition 2", right_on="Condition 2")
        relation.columns = ["S", "NS", "WT"]
    except:
        relation = pd.DataFrame({})
    samples_df = pd.DataFrame(samples)
    return relation, samples_df


def parse_params(param1, param2, output, main_log):
    """
    Given two parameter sheets, find samples in common and output
    df for later merging
    :param param1: parameter sheet for the first run
    :param param2: parameter sheet for the second run
    :param main_log: logger
    :return map_df: df with two cols [samples_1, samples_2]
    """
    with open(param1, "r") as p1:
        with open(param2, "r") as p2:
            param_one = json.load(p1)
            param_two = json.load(p2)
            wt_relation1, samples1 = read_json(param_one)
            wt_relation2, samples2 = read_json(param_two)

    # from these two files find samples in the same tile, with the same condition names
    # same replicates and tp
    # log how many samples will be merged
    merge = pd.merge(samples1, samples2, how="inner", on=["Tile ID", "Condition", "Time point", "Replicate"],
                     suffixes=["_run1", "_run2"])
    unique_Tid = merge["Tile ID"].drop_duplicates().tolist()
    unique_Conditions = merge["Condition"].drop_duplicates().tolist()
    merge["new_samples"] = merge["Sample ID_run1"] + "-" + merge["Sample ID_run2"]
    # save new sample names to output
    merge[["new_samples", "Tile ID", "Condition", "Time point", "Replicate"]].to_csv(os.path.join(output, "new_sample_names.csv"), index=False)
    main_log.info(f"{merge.shape[0]} samples will be merged")
    main_log.info(f"Tiles: {unique_Tid}")
    main_log.info(f"Condition: {unique_Conditions}")

    merge_wt = pd.concat([wt_relation1, wt_relation2]).drop_duplicates()
    return merge, merge_wt


def subtract_wt(input_dir1, input_dir2, map_df, wt_map, output):
    """

    """
    for index, row in map_df.iterrows():
        mut_count_f1 = os.path.join(input_dir1, f"counts_sample_{row['Sample ID_run1']}.csv")
        mut_count_f2 = os.path.join(input_dir2, f"counts_sample_{row['Sample ID_run2']}.csv")
        # process select, nonselect and wt together for each sample
        if row["Condition"] in wt_map["S"].tolist():
            print(row["Condition"])
            ns_name = wt_map[wt_map["S"] == row["Condition"]]["NS"].tolist()[0]
            wt_name = wt_map[wt_map["S"] == row["Condition"]]["WT"].tolist()[0]
            # corresponded wt and non-select
            wt = map_df.loc[(map_df["Tile ID"] == row["Tile ID"]) & (map_df["Time point"] == row["Time point"]) & (map_df["Replicate"] == row["Replicate"])]
            wt_run1_sample = wt[wt["Condition"] == wt_name]["Sample ID_run1"].tolist()[0]
            wt_run2_sample = wt[wt["Condition"] == wt_name]["Sample ID_run2"].tolist()[0]
            mut_count_wt1 = os.path.join(input_dir1, f"counts_sample_{wt_run1_sample}.csv")
            mut_count_wt2 = os.path.join(input_dir2, f"counts_sample_{wt_run2_sample}.csv")

            ns = map_df.loc[(map_df["Tile ID"] == row["Tile ID"]) & (map_df["Time point"] == row["Time point"]) & (
                        map_df["Replicate"] == row["Replicate"])]
            ns_run1_sample = ns[ns["Condition"] == ns_name]["Sample ID_run1"].tolist()[0]
            ns_run2_sample = ns[ns["Condition"] == ns_name]["Sample ID_run2"].tolist()[0]
            mut_count_ns1 = os.path.join(input_dir1, f"counts_sample_{ns_run1_sample}.csv")
            mut_count_ns2 = os.path.join(input_dir2, f"counts_sample_{ns_run2_sample}.csv")

            # remove wt counts from run 1
            updated_counts_run1, updated_counts_run1_ns,  updated_wt_run1 = remove_wt_counts(mut_count_f1,
                                                                                            mut_count_wt1, mut_count_ns1)
            updated_counts_run2, updated_counts_run2_ns, updated_wt_run2 = remove_wt_counts(mut_count_f2,
                                                                                            mut_count_wt2, mut_count_ns2)

            # find corresponded merged files and append df to the file
            mut_count_file = os.path.join(output, f"counts_sample_{row['Sample ID_run1']}-{row['Sample ID_run2']}.csv")
            mut_count_wt_file = os.path.join(output, f"counts_sample_{wt_run1_sample}-{wt_run2_sample}.csv")
            mut_count_ns_file = os.path.join(output, f"counts_sample_{ns_run1_sample}-{ns_run2_sample}.csv")

            # write mutations to selection condition
            merge = [updated_counts_run1, updated_counts_run2]
            merge = pd.concat(merge)
            merge_counts = merge.groupby(by=["HGVS"], as_index=False)["count"].sum().sort_values("count",
                                                                                                 ascending=False)
            merge_counts["count"] = merge_counts["count"].astype(int)
            merge_counts.to_csv(mut_count_file, mode="a", index=False)

            # write mutations to wt condition
            merge = [updated_wt_run1, updated_wt_run2]
            merge = pd.concat(merge)
            merge_counts = merge.groupby(by=["HGVS"], as_index=False)["count"].sum().sort_values("count",
                                                                                                 ascending=False)
            merge_counts["count"] = merge_counts["count"].astype(int)
            merge_counts.to_csv(mut_count_wt_file, mode="a", index=False)

            # write mutations to ns condition
            merge = [updated_counts_run1_ns, updated_counts_run2_ns]
            merge = pd.concat(merge)
            merge_counts = merge.groupby(by=["HGVS"], as_index=False)["count"].sum().sort_values("count",
                                                                                                 ascending=False)
            merge_counts["count"] = merge_counts["count"].astype(int)
            merge_counts.to_csv(mut_count_ns_file, mode="a", index=False)


def write_header(input_dir1, input_dir2, samples_df, output):
    """

    """
    new_samples = []
    for index, row in samples_df.iterrows():
        merged_header = ""
        mut_count_f1 = os.path.join(input_dir1, f"counts_sample_{row['Sample ID_run1']}.csv")
        mut_count_f2 = os.path.join(input_dir2, f"counts_sample_{row['Sample ID_run2']}.csv")
        new_samples.append(f"{row['Sample ID_run1']}-{row['Sample ID_run2']}")
        # join headers
        with open(mut_count_f1) as file1, open(mut_count_f2) as file2:
            for line1, line2 in zip(file1, file2):
                if "#" not in line1: break
                if "Final read-depth:" in line1:
                    n_depth_1 = int(line1.split(":")[1])
                    n_depth_2 = int(line2.split(":")[1])
                    total_depth = n_depth_1 + n_depth_2
                    merged_header += f"#Final read-depth:{total_depth}\n"
                    continue
                if "Number of read pairs without mutations:" in line1:
                    n_depth_1 = int(line1.split(":")[1])
                    n_depth_2 = int(line2.split(":")[1])
                    total_depth = n_depth_1 + n_depth_2
                    merged_header += f"#Number of read pairs without mutations:{total_depth}\n"
                    continue
                if "Total read pairs with mutations:" in line1:
                    n_depth_1 = int(line1.split(":")[1])
                    n_depth_2 = int(line2.split(":")[1])
                    total_depth = n_depth_1 + n_depth_2
                    merged_header += f"#Total read pairs with mutations:{total_depth}\n"
                    continue
                merged_header += line1.strip() + "; " + line2
        output_f = open(os.path.join(output, f"counts_sample_{row['Sample ID_run1']}-{row['Sample ID_run2']}.csv"), "w")
        output_f.write(merged_header)
        output_f.close()


def remove_wt_counts(mut_count_f, wt_f, ns_f):
    """

    """
    # load  df
    # skip header and read the mutations
    df = pd.read_csv(mut_count_f, skiprows=18)
    # load wt df
    # skip header and read the mutations
    wt_df = pd.read_csv(wt_f, skiprows=18)
    # load ns df
    # skip header and read the mutations
    ns_df = pd.read_csv(ns_f, skiprows=18)

    # from non-select, remove wt counts
    merged_df_ns = pd.merge(ns_df, wt_df, on="HGVS", how="outer", suffixes=["_ns", "_wt"])
    merged_df_ns = merged_df_ns.fillna(0)
    # subtract wt
    merged_df_ns["updated_ns"] = merged_df_ns["count_ns"] - merged_df_ns["count_wt"]

    # for all the values <= 0, assign 0 to the updated wt
    merged_df_ns["updated_wt"] = merged_df_ns["count_wt"]

    merged_df_ns.loc[(merged_df_ns["updated_ns"] <= 0) & (merged_df_ns["count_ns"] != 0), "updated_wt"] = 0
    merged_df_ns.loc[(merged_df_ns["updated_ns"] <= merged_df_ns["count_ns"]) & (merged_df_ns["updated_ns"] > 0),
                     "updated_wt"] = 0

    merged_df_ns.loc[(merged_df_ns["updated_ns"] <= 0) & (merged_df_ns["count_ns"] > 0), "updated_ns"] = 0
    merged_df_ns.loc[merged_df_ns["count_ns"] == 0, "updated_ns"] = 0
    updated_counts_ns = merged_df_ns.loc[merged_df_ns["updated_ns"] != 0][["HGVS", "updated_ns"]]
    updated_counts_ns.columns = ["HGVS", "count"]
    updated_counts_wt_ns = merged_df_ns.loc[merged_df_ns["updated_wt"] != 0][["HGVS", "updated_wt"]]
    updated_counts_wt_ns.columns = ["HGVS", "ns_count"]

    # from select, remove wt counts
    merged_df = pd.merge(df, wt_df, on="HGVS", how="outer", suffixes=["_s", "_wt"])
    merged_df = merged_df.fillna(0)
    # subtract wt
    merged_df["updated_s"] = merged_df["count_s"] - merged_df["count_wt"]

    # for all the values <= 0, assign 0 to the updated wt
    merged_df["updated_wt"] = merged_df["count_wt"]

    merged_df.loc[(merged_df["updated_s"] <= 0) & (merged_df["count_s"] != 0), "updated_wt"] = 0
    merged_df.loc[(merged_df["updated_s"] <= merged_df["count_s"]) & (merged_df["updated_s"] > 0), "updated_wt"] = 0

    merged_df.loc[(merged_df["updated_s"] <= 0) & (merged_df["count_s"] > 0), "updated_s"] = 0
    merged_df.loc[merged_df["count_s"] == 0, "updated_s"] = 0
    updated_counts = merged_df.loc[merged_df["updated_s"] != 0][["HGVS", "updated_s"]]
    updated_counts.columns = ["HGVS", "count"]

    updated_counts_wt_s = merged_df.loc[merged_df["updated_wt"] != 0][["HGVS", "updated_wt"]]
    updated_counts_wt_s.columns = ["HGVS", "count"]

    # join updated counts wt ns and s into new wt
    # the updated wt counts contains the variants that are not found in ns or s
    updated_counts_wt = pd.merge(updated_counts_wt_ns, updated_counts_wt_s, how="inner", on="HGVS")
    return updated_counts, updated_counts_ns, updated_counts_wt


def merge_main(args):

    # check args and make corresponding output dir
    # log input/output information
    if not os.path.isdir(args.output):
        os.mkdir(args.output)
    logging = logginginit(args.log_level, os.path.join(args.output, "mergeRuns.log"))
    main_log = logging.getLogger("mergeRuns.log")
    main_log.info(f"Input dir 1: {args.dir1} with Parameter Sheet: {args.paramOne}")
    main_log.info(f"Input dir 2: {args.dir2} with Parameter Sheet: {args.paramTwo}")
    main_log.info(f"Merges mutation count file will be saved to {args.output}")
    map_df, wt_map_df = parse_params(args.paramOne, args.paramTwo, args.output, main_log)
    if args.subtractWT:
        main_log.info("subtract WT turned on, wt counts will be subtracted from select/non-select conditions")
        write_header(args.dir1, args.dir2, map_df, args.output)
        subtract_wt(args.dir1, args.dir2, map_df, wt_map_df, args.output)

    else:
        # map_df = parse_params(args.paramOne, args.paramTwo, args.output, main_log)
        new_samples = merge_runs(args.dir1, args.dir2, map_df, args.output)

    main_log.info(f"Merged files are saved to {args.output}")

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='TileSeq mutation counts')
    # user input arguments
    parser.add_argument("-p1", "--paramOne", help="Path to parameter sheet (JSON) for the first run", type=str, required=True)
    parser.add_argument("-p2", "--paramTwo", help="Path to parameter sheet (JSON) for the second run", type=str, required=True)
    parser.add_argument("-d1", "--dir1", help="Path to *_mut_count folder for the first run", type=str, required=True)
    parser.add_argument("-d2", "--dir2", help="Path to *_mut_count folder for the second run", type=str, required=True)
    parser.add_argument("-o", "--output", help="Output folder path (this folder will be made if not exists when "
                                               "running the script",type=str, required=True)
    parser.add_argument("--subtractWT", help="Subtract wt counts from non-select before merging", action="store_true")
    parser.add_argument("-log", "--log_level", help="set log level: debug, \
            info, warning, error, critical. (default = debug)", type=str, default="debug")
    # test dirs (local, CHEK2)
    # dir r2r3 /Users/roujia/Desktop/Tileseq_DMS/CHEK2/210216_CHEK2_R2R3/210216_CHEK2_R2R3_2021-02-23-18-39-16_mut_count
    # dir 200310 /Users/roujia/Desktop/Tileseq_DMS/CHEK2/200310_rerun_CHEK2/v419_2021-02-02-17-57-49_mut_count
    # param1 210216_CHEK2.json
    # param2 200310_CHECK2_newtile11.json

    args = parser.parse_args()

    merge_main(args)



#!/usr/bin/env python3.7

# Helper functions
# 1. downsample fastq files into n reads
# 2. loading process animation function
# 3. parse input json file

import pandas as pd
import random
import os
import sys
import json
import time
import itertools
import logging
from Bio.Seq import Seq

def downsample(n, r1, r2, output_path):
    """
    Randomly downsample r1 and r2 to n reads
    """

    record_number = 0

    base_r1 = os.path.basename(r1)
    base_r2 = os.path.basename(r2)

    output_r1 = output_path+"/"+base_r1.replace(".fastq", "_ds.fastq")
    output_r2 = output_path+"/"+base_r2.replace(".fastq", "_ds.fastq")

    with open(r1, "r") as i:
        r1_lines = sum([1 for line in i])
    with open(r2, "r") as j:
        r2_lines = sum([1 for line in j])

    total_records_r1 = int(r1_lines/4)
    total_records_r2 = int(r2_lines/4)

    if n > total_records_r1 or n > total_records_r2:
        os.system("cp "+r1+ " "+output_r1)
        os.system("cp "+r2+ " "+output_r2)
        return output_r1, output_r2
    else:
        if total_records_r1 <= total_records_r2:
            records_to_keep = set(random.sample(range(total_records_r1 + 1), n))
        else:
            records_to_keep = set(random.sample(range(total_records_r2 + 1), n))

    with open(r1, "r") as i_1:
        with open(output_r1, "w") as o1:

            for line1 in i_1:
                line2 = i_1.readline()
                line3 = i_1.readline()
                line4 = i_1.readline()
                if record_number in records_to_keep:
                    o1.write(line1)
                    o1.write(line2)
                    o1.write(line3)
                    o1.write(line4)
                record_number += 1


    record_number = 0
    with open(r2, "r") as i_2:
        with open(output_r2, "w") as o2:
            for line1 in i_2:
                line2 = i_2.readline()
                line3 = i_2.readline()
                line4 = i_2.readline()
                if record_number in records_to_keep:
                    o2.write(line1)
                    o2.write(line2)
                    o2.write(line3)
                    o2.write(line4)
                record_number += 1

    return output_r1, output_r2

def parse_json(json_file):
    """
    parse input json file
    return information about the run
    """
    with open(json_file, "r") as jsonf:
        data = json.load(jsonf)
        project = data["project"] # project name
        seq = pd.DataFrame(data["template"], index=[0]) # sequence
        cds_seq = seq.seq[0]
        cds_seq = cds_seq[int(seq.cds_start)-1: int(seq.cds_end)]
        # create dictionary to map tile/region # to start,end positions
        try:
            tile_map = pd.DataFrame(data["tiles"])
        except:
            tile_map = pd.DataFrame(data["tiles"], index=[0])

        try:
            region_map = pd.DataFrame(data["regions"])
        except:
            region_map = pd.DataFrame(data["regions"], index=[0])
        try:
            samples = pd.DataFrame.from_dict(data["samples"])
        except:
            samples = pd.DataFrame.from_dict(data["samples"], index=[0])

        # todo get relationship table for future use
        # todo special case: no WT
        # todo special case: all WT
        relations = data["conditions"]["definitions"]
        # get posterior quality cut off
        var_caller = data["varcaller"]
    return project, seq, cds_seq, tile_map, region_map, samples, var_caller, relations


def loadingAnimation(process):
    """
    loading animation to show when loading a process
    """

    while process.is_alive():
        chars = "/—\|"
        for char in chars:
            sys.stdout.write('\r'+'downsamping reads ... '+char)
            time.sleep(.1)
            sys.stdout.flush()
    sys.stdout.write("\n")


def build_lookup(cds_start, cds_end, cds_seq):
    """
    build a lookup df for a given gene
    return a df with columns:
    temp_pos, na, dna_pos, protein, protein_pos
    """
    lookup_table = {}
    # list of template pos
    temp = list(range(cds_start, cds_end+1))
    # list of coding DNA bases
    cDNA = [i for i in cds_seq]
    cDNA_pos = list(range(1, len(cDNA)+1))
    # list of protein bases
    protein = [i for i in Seq(cds_seq).translate()]
    protein_pos = range(1, len(protein)+1)
    lookup_table["temp_pos"] = temp
    lookup_table["cds"] = cDNA
    lookup_table["cds_pos"] = cDNA_pos
    lookup_table["protein"] = list(itertools.chain.from_iterable(itertools.repeat(x, 3) for x in protein))
    lookup_table["protein_pos"] = list(itertools.chain.from_iterable(itertools.repeat(x, 3) for x in protein_pos))
    df = pd.DataFrame(lookup_table)
    return df


def logginginit(log_level, log_f):
    """
    Init logging in console as well as main log
    """
    log_level = log_level.upper()
    logging.basicConfig(filename=log_f,
                        filemode="w",
                        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
                        datefmt="%m/%d/%Y %I:%M:%S %p",
                        level=log_level)

    # define a Handler which writes INFO messages or higher to the sys.stdout
    console = logging.StreamHandler(sys.stdout)
    console.setLevel(log_level)
    # set a format which is simpler for console use
    formatter = logging.Formatter('%(asctime)s - %(name)-8s: %(levelname)-4s %(message)s')
    # tell the handler to use this format
    console.setFormatter(formatter)
    # add the handler to the root logger
    logging.getLogger('').addHandler(console)

    # stderr_logger = logging.getLogger('STDERR')
    # sl = StreamToLogger(stderr_logger, logging.ERROR)
    # sys.stderr = sl

    return logging

class StreamToLogger(object):
   """
   Fake file-like stream object that redirects writes to a logger instance.
   """
   def __init__(self, logger, log_level=logging.INFO):
      self.logger = logger
      self.log_level = log_level
      self.linebuf = ''

   def write(self, buf):
      for line in buf.rstrip().splitlines():
         self.logger.log(self.log_level, line.rstrip())


if __name__ == "__main__":
    j = "/home/rothlab2/rli/tileseq_output/210416_CHEK2_QC_2021-04-19-16-02-03/210416_optimizedRef_CHEK2.json"
    j = "/home/rothlab2/rli/tileseq_output/210120_LDLuptake_2021-01-25-13-13-43/Parameter-sheet_LDLR_20210120_LDLuptake.json"
    parse_json(j)



#!/usr/bin/env python3.7

# Main script for sequencing analysis
# Input of this script:
#        - JSON file containing input parameters
#    - Fastq files (gzipped)

# what does this script do?
# 1. Read input csv file and convert to json (using csv2Json.R (you need to
# install required packages in R)
# 1. Read paramters from json file
# 2. Read input fastq file
#    2.a. Ramdonly select 30k reads and save a copy of downsampled fastq files

# 3. Align fastq file to reference sequence, generate sam files
# 4. From sam files, count mutations
# 5. Output mutation counts to summary.csv

# modules
import re
import subprocess

import pandas as pd
import numpy as np
import os
import glob
import argparse
import logging
import datetime
import shutil
import time
from pathlib import Path

# pakage modules
from TileSeqMut import help_functions
from TileSeqMut import count_mut
from TileSeqMut import cluster

# pakage modules
# import help_functions
# import count_mut
# import cluster

class fastq2counts(object):
    """Main class to read fastq files and output mutation counts."""

    def __init__(self, param_path, output_dir, main_log, args):
        """
        Initialize mutation counts
        param_path: Full path to param.json file
        output_dir: Directory to save all the output files
        main_log: logging object
        args: user inpu arguments
        """
        self._main_path = os.path.abspath(__file__)  # path to this python file
        self._logging = main_log
        self._param_json = param_path  # parameter json file

        self._args = args  # user input arguments
        self._args.environment = self._args.environment.upper()

        if self._args.fastq:
            self._fastq_list = glob.glob(self._args.fastq+"/*.fastq.gz")
        else:
            self._fastq_list = []

        self._output = output_dir

        # parse parameter json file
        self._project, self._seq, self._cds_seq, self._tile_map, self._region_map, \
            self._samples, self._var, self._relations = help_functions.parse_json(param_path)
        self._sample_names = self._samples["Sample ID"].tolist()
        self._project = self._project.replace(" ", "_")  # project name
        # make main log
        self._log = self._logging.getLogger("main.log")
        self._log.debug(self._args.environment)

    def _init_skip(self, skip, r1=None, r2=None):
        """
        if skip == T: skip alignment
            make suboutput dir with time stamp
        else: not skip alignment
            make main output dir with time stamp
        """
        if skip:  # only run mutation counts on the aligned samples
            self._skip = True
            if self._args.r1 and self._args.r2:
                self._r1 = self._args.r1
                self._r2 = self._args.r2
                # self._log.info(f"SAM R1: {r1}")
                # self._log.info(f"SAM R2: {r2}")
            else:
                self._r1 = ""
                self._r2 = ""
                self._log.info(f"Sam files are read from {self._output}")
        else:
            self._r1 = ""
            self._r2 = ""
            self._skip = False

    def _align_sh_(self):
        """
        1. Align each fastq file to reference (submit one job for each alignment)
        2. Align each downsampled fastq file to reference (submit one job for each alignment)
        Output sam files into output folders (sam_files, ds_sam_files)
        """
        # VALIDATE: if all samples are present in fastq files
        # also check if any fastq file is empty
        fastq_sample_id = [os.path.basename(i).split("_")[0] for i in self._fastq_list]
        fastq_sample_id = list(set(fastq_sample_id))
        self._log.debug(f"Fastq sample ID in input folder: {fastq_sample_id}")
        # validation
        # check if input fastq files contains all the samples in paramter file
        self._log.debug(f"Sample names in parameter sheet: {set(self._sample_names)}")
        if set(self._sample_names).issubset(fastq_sample_id):
            self._log.info("All samples found")
            self._log.info(f"In total there are {len(list(set(self._sample_names)))} samples in the csv file")
            self._log.info(f"In total there are {len(fastq_sample_id)} fastq files")
        else:
            test= list(np.setdiff1d(self._sample_names,fastq_sample_id))
            join_list = ",".join(test)
            self._log.error("fastq files do not match input samples.")
            self._log.error(f"Fastq files not found for sample {join_list}")
            raise FileNotFoundError()


        # create mappving for R1 and R2 for each sample
        # ONLY samples provided in the parameter files will be analyzed
        fastq_map = []
        for r1 in self._fastq_list:
            ID = os.path.basename(r1).split("_")[0]
            if ("_R1_" in r1) and (ID in self._sample_names):
                    r2 = r1.replace("_R1_", "_R2_")
                    fastq_map.append([r1,r2])

        # make folder to store alignment sam files
        sam_output = os.path.join(self._output, "sam_files/")
        os.system("mkdir " + sam_output)

        # make folder to store ref
        ref_path = os.path.join(self._output, "ref/")
        os.system("mkdir " + ref_path)

        # convert to df
        fastq_map = pd.DataFrame(fastq_map, columns=["R1", "R2"])
        # if phix
        if self._args.calibratePhredPhix:
            # check if Undetermined reads are in the same folder
            self._phix_fastq = glob.glob(self._args.fastq+"/Undetermined_*.fastq.gz")
            if len(self._phix_fastq) < 2:
                raise ValueError("Cannot find Undetermined fastq files for phix alignment")
        if self._phix_fastq != []:
            fastq_map.loc[len(fastq_map)] = self._phix_fastq
            phix_fasta = os.path.join(os.path.dirname(self._main_path), "data/phix.fasta")
            #self._log.info(os.path.dirname(self._main_path))
            # save this to ref path
            #self._log.info(phix_fasta)
            cmd = f"cp {phix_fasta} {ref_path}"
            #self._log.info(cmd)
            os.system(cmd)

        rc = False
        if self._args.rc:
            rc = True
        # GURU
        if self._args.environment == "GURU":
            # make sh files to submit to GURU
            sh_output = os.path.join(self._output, "GURU_jobs")
            os.system("mkdir "+sh_output)

            # for each pair of fastq files, submit jobs to run alignment
            # it takes the following arguments: R1, R2, ref, Output sam
            # the output log (bowtie log) would be in the same dir
            logging.info("Writing sh files for alignment (GURU)")
            cluster.alignment_sh_guru(fastq_map, self._project, self._seq.seq.item(), ref_path, sam_output, sh_output, logging)
            logging.info("Alignment jobs are submitted to GURU..")

            # wait for alignment to finish and call mutations
        elif self._args.environment == "GALEN":
            # make sh files to submit to GURU
            sh_output = os.path.join(self._output, "GALEN_jobs")
            os.system("mkdir "+sh_output)

            # for each pair of fastq files, submit jobs to run alignment
            # it takes the following arguments: R1, R2, ref, Output sam
            # the output log (bowtie log) would be in the same dir
            logging.info("Writing sh files for alignment (GALEN)")
            sam_df, job_list = cluster.alignment_sh_galen(fastq_map, self._project, self._seq.seq.values.item(), ref_path,
                                            sam_output, sh_output, self._args.at, self._log, rc)
            logging.info("Alignment jobs are submitted to GALEN..")

        elif self._args.environment == "BC2" or self._args.environment == "BC":
            # make sh files to submit to BC
            sh_output = os.path.join(self._output, "BC_aln_sh")
            os.system("mkdir " + sh_output)
            # make sh files to submit to BC
            self._log.info("Submitting alignment jobs to BC/BC2...")
            # alignment_sh_bc2(fastq_map, ref_name, ref_seq, ref_path, sam_path, sh_output, at, logging)

            sam_df, job_list = cluster.alignment_sh_bc2(fastq_map, self._project, self._seq.seq.values.item(),
                                                        ref_path, sam_output, sh_output, self._args.at,
                                                        self._log, rc)
            self._log.info("Alignment jobs are submitte to BC2. Check pbs-output for STDOUT/STDERR")

        elif self._args.environment == "DC":
            # make sh files to submit to DC
            sh_output = os.path.join(self._output, "DC_aln_sh")
            os.system("mkdir "+sh_output)

            self._log.info("Submitting alignment jobs to DC...")
            if  self._args.rc:
                rc = True
            else:
                rc = False
            sam_df, job_list = cluster.alignment_sh_dc(fastq_map, self._project, self._seq.seq.values.item(),
                                                       ref_path, sam_output, sh_output,
                                                       self._args.at, self._log, rc)
            self._log.info("Alignment jobs are submitte to DC. Check pbs-output for STDOUT/STDERR")

        else:
            raise ValueError("Wrong environment name")

        self._log.info(f"Total jobs submitted: {len(job_list)}")
        if self._args.environment == 'GALEN':
            finished = cluster.parse_jobs_galen(job_list, self._logging.getLogger("track.jobs"))
        else:
            finished = cluster.parse_jobs(job_list, self._args.environment, self._logging.getLogger("track.jobs"))  #
        # track list
        # of jobs

        if finished:

            self._log.info(f"Alignment jobs are finished!")
            self._log.info(f"Merging alignment log files...")
            # go through alignment output dir and parse all the log files
            log_files = os.path.join(sam_output, "*.log")
            log_list = glob.glob(log_files)
            alignment_master_log = []
            for f in log_list:
                f_info = []
                sample_id = os.path.basename(f).split(".")[0]
                f_info.append(sample_id)
                with open(f, "r") as fp:
                    for i, line in enumerate(fp):
                        # pick corresponding lines in the alignment log file
                        if i == 0 or i == 5 or i == 6 or i == 11:
                            line = line.split(" ")
                            f_info.append(line[0])
                alignment_master_log.append(f_info)
            df = pd.DataFrame(alignment_master_log)
            df.columns = ["sample_id", "R1_reads", "R1_alignment_rate", "R2_reads", "R2_alignment_rate"]
            alignment_log = os.path.join(sam_output, "aligned_rate.log")
            df.to_csv(alignment_log, index=False)
            self._log.info(f"Alignment log file created")

        return sam_output

    def _mut_count(self):
        """
        Count mutations in input sam files
        1. Log path to sam files and downsampled sam files
        2. If sam df not provided, make df for all the sam files
        3. For each pair of sam files, submit jobs to the cluster
        4. Log to main log
        """
        self._log.info("Counting mutations ...")

        # submit job with main.py -r1 and -r2
        # run main.py with -r1 and -r2
        logger_mut = self._logging.getLogger("count.mut.log")
        mut_counts = count_mut.readSam(self._r1, self._r2, self._param_json, self._args, self._output, self._args.c, logger_mut)
        self._log.info("Running multi-core analysis ... ")
        # todo add adjust error rate here
        if self._args.calibratePhredWT:
            self._log.info("Adjusting phred scores using WT samples...")
            adjusted_phred = mut_counts.adjust_er(wt_override=self._args.wt_override)
        elif self._args.calibratePhredPhix:
            self._log.info("Adjusting phred scores using Phix...")
            adjusted_phred = mut_counts.adjust_er_phix()
        else:
            adjusted_phred = []
        mut_counts.multi_core(adjusted_phred)
        # end = time.time()
        # print('Time taken for 8 cores program: ', end - start)

    def _makejobs(self, sh_output, sam_dir, samples=""):
        """
        For each pair of sam files in output/sam_files/
        submit mut count job to the cluster
        """
        if samples == "":
            samples = self._sample_names


        if sam_dir == "":
            # get samples in param file
            sam_dir = os.path.join(self._args.output, "sam_files/") # read sam file from sam_file
        # get sam files from parameter file
        if not os.path.isdir(sam_dir):
            self._log.error(f"Directory: ./sam_files/ not found in {self._output}")
            raise ValueError()

        #if self._args.calibratePhredPhix:
        #    samples.append("Undetermined")

        self._log.debug(f"Sam files are read from {sam_dir}")
        job_list = []
        for sample in samples:
            # assume all the sam files have the same name format (id_*.sam)
            sam_f_r1 = glob.glob(f"{sam_dir}/{sample}_*R1_*.sam")
            sam_f_r2 = glob.glob(f"{sam_dir}/{sample}_*R2_*.sam")
            if len(sam_f_r1) == 0 or len(sam_f_r2) == 0:
                self._log.error(f"SAM file for sample {sample} not found. Please check your parameter file")
                raise ValueError()

            else:
                self._log.info(f"Sample {sample}")
                self._log.info(f"Read1: {sam_f_r1[0]}")
                self._log.info(f"Read2: {sam_f_r2[0]}")
                sam_id_1 = os.path.basename(sam_f_r1[0]).split("_")[0]
                sam_id_2 = os.path.basename(sam_f_r2[0]).split("_")[0]
                if (sam_id_1 != sample) or (sam_id_1 != sam_id_2) or (sam_id_2 != sample):
                    self._log.error("IDs in sam files don't match!")
                self._r1 = sam_f_r1[0]
                self._r2 = sam_f_r2[0]
            
            job_id = self._submit_mutcount_jobs(sample, sh_output)
            # submit job with main.py -r1 and -r2
            # run main.py with -r1 and -r2
            # if self._args.sr_Override:
            #     cmd = f"tileseq_mut -n {self._args.name} -r1 {self._r1} -r2 {self._r2} -o {self._output} -p" \
            #           f" {self._param_json} --skip_alignment -log {self._args.log_level} -env {self._args.environment} -at {self._args.at} -mt {self._args.mt} -override"
            # else:
            #     cmd = f"tileseq_mut -n {self._args.name} -r1 {self._r1} -r2 {self._r2} -o {self._output} -p {self._param_json} --skip_alignment -log {self._args.log_level} -env {self._args.environment} -at {self._args.at} -mt {self._args.mt}"
            #
            # if self._args.wt_override:
            #     cmd = f"tileseq_mut -n {self._args.name} -r1 {self._r1} -r2 {self._r2} -o {self._output} -p" \
            #           f" {self._param_json} --skip_alignment -log {self._args.log_level} -env {self._args.environment} -at {self._args.at} -mt {self._args.mt} --wt_override"
            # else:
            #     cmd = f"tileseq_mut -n {self._args.name} -r1 {self._r1} -r2 {self._r2} -o {self._output} -p {self._param_json} --skip_alignment -log {self._args.log_level} -env {self._args.environment} -at {self._args.at} -mt {self._args.mt}"
            #
            #
            #
            # if self._args.environment == "BC2" or self._args.environment == "BC":
            #     logging.info("Submitting mutation counts jobs to BC2...")
            #     job_id = cluster.mut_count_sh_bc(sample, cmd, self._args.mt, self._args.mm, sh_output, self._log,
            #                                      self._args.c)
            # elif self._args.environment == "DC":
            #     logging.info("Submitting mutation counts jobs to DC...")
            #     # (sample_name, cmd, mt, sh_output_dir, logger)
            #     job_id = cluster.mut_count_sh_dc(sample, cmd, self._args.mt, self._args.mm, sh_output, self._log,
            #                                      self._args.c) # this
            #     # function
            #     # will make a sh file for submitting the job
            # elif self._args.environment == "GALEN":
            #     logging.info("Submitting mutation counts jobs to GALEN...")
            #     # (sample_name, cmd, mt, sh_output_dir, logger)
            #     # todo finish this function
            #     job_id = cluster.mut_count_sh_galen(sample, cmd, self._args.mt, self._args.mm, sh_output, self._log,
            #                                      self._args.c) # this
            # else:
            #     raise ValueError()

            job_list.append(job_id)
        self._log.info("all samples submitted")

        jobs = ",".join(job_list)
        self._log.debug(f"All jobs: {jobs}")
        self._log.info(f"Total jobs running: {len(job_list)}")
        if self._args.environment == 'GALEN':
            finished = cluster.parse_jobs_galen(job_list, self._logging.getLogger("track.jobs"))
        else:
            finished = cluster.parse_jobs(job_list, self._args.environment, self._logging.getLogger("track.jobs")) # track
        # list of jobs

        return finished

    def _submit_mutcount_jobs(self, sample, sh_output):
        """
        Make mutation counting jobs and submit for this sample
        """
        # submit job with main.py -r1 and -r2
        # run main.py with -r1 and -r2
        cmd = f"tileseq_mut -n {self._args.name} -r1 {self._r1} -r2 {self._r2} -o {self._output} -p" \
                  f" {self._param_json} --skip_alignment -log {self._args.log_level} -env {self._args.environment} -at {self._args.at} -mt {self._args.mt} " 
        if self._args.sr_Override:
            cmd = cmd + "-override "

        if self._args.wt_override:
            cmd = cmd + "--wt_override "
        
        if self._args.calibratePhredPhix:
            cmd = cmd + "--calibratePhredPhix "

        if self._args.calibratePhredWT:
            cmd = cmd + "--calibratePhredWT "

        if self._args.environment == "BC2" or self._args.environment == "BC":
            logging.info("Submitting mutation counts jobs to BC2...")
            job_id = cluster.mut_count_sh_bc(sample, cmd, self._args.mt, self._args.mm, sh_output, self._log,
                                             self._args.c)
        elif self._args.environment == "DC":
            logging.info("Submitting mutation counts jobs to DC...")
            # (sample_name, cmd, mt, sh_output_dir, logger)
            job_id = cluster.mut_count_sh_dc(sample, cmd, self._args.mt, self._args.mm, sh_output, self._log,
                                             self._args.c)  # this
            # function
            # will make a sh file for submitting the job
        elif self._args.environment == "GALEN":
            logging.info("Submitting mutation counts jobs to GALEN...")
            # (sample_name, cmd, mt, sh_output_dir, logger)
            job_id = cluster.mut_count_sh_galen(sample, cmd, self._args.mt, self._args.mm, sh_output, self._log,
                                                self._args.c)  # this
        else:
            raise ValueError("Wrong environment")

        return job_id

    def _checkoutput(self, f):
        """
        Check how many variants are generated in each file
        """
        mut_n = 0
        with open(f, "r") as mut_output:
            for line in mut_output:
                # skip header
                if "c." in line:
                    mut_n += 1
        return mut_n

    def _main(self):
        """
        Main for the fastq2counts object
        """
        sam_dir = ""
        if self._skip == False: # submit jobs for alignment
            # get sam_df from self._align_sh()
            sam_dir = self._align_sh_()

        if self._r1 != "" and self._r2 != "":
            # call functions in count_mut.py
            # init(sam_r1, sam_r2, param, args, ouptut_dir, logger)
            self._mut_count()
        # submit jobs for mutation counting
        # if user did not provide r1 and r2 SAM file
        # get r1 and r2 sam files from output dir provided by user
        elif self._r1 == "" and self._r2 == "":

            if self._args.resubmit:
                # resubmit failed jobs in existing mut_count dir
                # self.output is the mut_count dir in this case
                # find out which jobs failed by going through all counts files
                mutcount_list = glob.glob(os.path.join(self._output, "counts_sample_*.csv"))
                failed_samples = []
                all_samples_with_csv = []
                for f in mutcount_list:
                    m = re.search('.*counts_sample_(.+?).csv', f)
                    if m:
                        sample_id = m.group(1)
                    else:
                        self._log.error("Sample ID not found")
                        raise ValueError(f"Sample ID not found for {f}")
                    all_samples_with_csv.append(sample_id)
                    mut_n = self._checkoutput(f)
                    if mut_n == 0:
                        failed_samples.append(sample_id)
                other_missing = [i for i in self._sample_names if i not in all_samples_with_csv]
                failed_samples += other_missing
                self._log.info(f"Failed samples: {failed_samples}")
                self._log.info("Resubmitting samples ...")
                # find the path to jobs dir
                env_jobs_dir = os.path.join(self._output, f"{self._args.environment}_jobs")
                parent_dir = os.path.abspath(os.path.join(self._output, os.pardir))
                sam_dir = os.path.join(parent_dir, "sam_files")
                # remake sh files for these samples and submit them
                finished = self._makejobs(env_jobs_dir, sam_dir, samples=failed_samples)
                
                self._log.info("ALL DONE!")
                return finished

            if self._skip == False:
                time_now = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
                # make mut count dir alignemnt not skipped
                # otherwise the user input dir should be the output dir gnerated
                self._output = os.path.join(self._output, self._args.name + "_" + time_now + "_mut_count")
                os.makedirs(self._output)
                # align to phix if specified
                if self._args.calibratePhredPhix:
                    # check if Undetermined reads are in the same folder
                    self._phix_fastq = glob.glob(self._args.fastq+"/Undetermined_*.fastq.gz")
                    if len(self._phix_fastq) < 2:
                        raise ValueError("Cannot find Undetermined fastq files for phix alignment")
                    else:
                        self._phix_fastq = []


            # output directory is the mut_count dir
            # make folder to store all the sh files as well as all the log files
            if self._args.environment == "BC2" or self._args.environment == "BC":
                sh_output = os.path.join(self._output, "BC_jobs")

            elif self._args.environment == "DC":
                sh_output = os.path.join(self._output, "DC_jobs")

            elif self._args.environment == "GALEN":
                sh_output = os.path.join(self._output, "GALEN_jobs")

            else:
                self._log.error("Please provide valid environment: BC/BC2/DC/Galen")
                raise ValueError("Please provide valid environment: BC/BC2/DC/Galen")

            self._log.info(f"Mutation count sh files are made in {sh_output}")

            os.mkdir(sh_output)
            finished = self._makejobs(sh_output, sam_dir)

            if finished:
                self._log.info(f"Mutation counting jobs are finished!")
                self._log.info("Check mutation counts file ...")
                mutcount_list = glob.glob(os.path.join(self._output, "counts_sample_*.csv"))
                self._log.info(f"{len(mutcount_list)} mutation counts file generated")
                if len(self._sample_names) > len(mutcount_list):
                    self._log.error("Job finished but some variant call files are missing!!")
                    missing = list(set(self._sample_names)-set(mutcount_list))
                    missing = ",".join(missing)
                    self._log.error(f"Missing samples: {missing}")
                for f in mutcount_list:
                    mut_n = self._checkoutput(f)
                    if mut_n == 0:
                        self._log.error(f"{f} has 0 variants! Check mut log for this sample.")
                    else:
                        self._log.info(f"{f} has {mut_n} variants")


def check(args):
    """
    Validate args and convert csv to JSON
    return path to json
    """
    # todo test if tileseqMave is installed correctly
    # two script needed: csv2json.R and calibratePhred.R
    test_csv2json = subprocess.getstatusoutput("csv2json.R -h")
    test_caliPhred = subprocess.getstatusoutput("calibratePhred.R -h")
    if test_csv2json[0] != 0:
        raise OSError("csv2json.R not installed or not found in path")
    if test_caliPhred[0] != 0:
        raise OSError("calibratePhred.R not installed or not found in path")

    # check if the correct args are provided
    # if you don't specify --skip_alignment then you cannot provide r1 and r2 sam_files
    if not args.skip_alignment:
        if args.r1 or args.r2:
            raise ValueError(f"Invalid paramters! Please specify --skip_alignment if you want to analyze\
            one pair of sam files")

    # check if output dir exists
    if not os.path.isdir(args.output):
        raise FileNotFoundError(f"Output directory not found: {args.output}")

    # check if fastq dir exists
    if args.fastq and not os.path.isdir(args.fastq):
        raise FileNotFoundError(f"Fastq file path not found: {args.fastq}")

    # try convert csv to json in the same dir as the csv file
    # convert csv file to json
    if args.param.endswith(".csv"):
        print("parameter sheet in csv format, converting to json")
        param_json = args.param.replace(".csv", ".json")
        if os.path.isfile(param_json):
             os.remove(param_json)
        if args.sr_Override:
            convert = f"csv2json.R {args.param} -o {param_json} --srOverride"
        else:
            convert = f"csv2json.R {args.param} -o {param_json}"
        os.system(convert)

    # if the file ends with .json, do nothing
    elif args.param.endswith(".json"):
        param_json = args.param
    else:
        raise ValueError("Please provide valid paramter file format (csv or json)")

    if not os.path.isfile(param_json):
        raise ValueError("Json file does not exist, check conversion!")

    # check environment input
    if args.environment.upper() not in ["BC", "BC2", "DC", "GALEN"]:
        raise ValueError(f"Wrong input for environment: {args.environment}")

    return param_json


def main(args):
    """
    Main for fastq2counts
    """
    # get time stamp for this object
    time_now = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")

    param_json = check(args)
    # get basename for the param file
    param_base = os.path.basename(param_json)
    if args.skip_alignment: # if we want to skip the alignment part
        # if user provided R1sam + R2sam
        if args.r1 and args.r2:
            # analyze one pair of r1 and r2 file
            # VALIDATE if both files are sam files
            if not args.r1.endswith(".sam") or not args.r2.endswith(".sam"):
                raise ValueError("Please provide SAM files")
            # copy the parameter file to output dir if the json file is not in the dir
            param_path = os.path.join(args.output, param_base)
            if not os.path.isfile(param_path):
                param_path = shutil.copy(param_json, args.output, follow_symlinks=True)
            # set up loggingthis creates main.log in the mut_count output dir
            main_log_f = os.path.join(args.output, "mutcount_main.log")
            main_log = help_functions.logginginit(args.log_level, main_log_f)

            # initialize mutcount object
            mc = fastq2counts(param_path, args.output, main_log, args)
            mc._init_skip(skip=True, r1=args.r1, r2=args.r2)

        else:
            if not args.resubmit:
            
                # user provided a csv file and path to sam files
                # for each pair of sam file we would submit one job to the cluster for mutation counting
                # make time stamped output folder for this project
                updated_out = os.path.join(args.output, args.name + "_" + time_now + "_mut_count")
                # sam_dir = os.path.join(args.output, "sam_files/") # read sam file from sam_file
                os.makedirs(updated_out)
            else:
                updated_out = args.output
            args_log_path = os.path.join(updated_out, "args.log")
            write_param(args_log_path, args)
            # load json file
            param_path = os.path.join(updated_out, param_base)
            if not os.path.isfile(param_path):
                param_path = shutil.copy(param_json, updated_out, follow_symlinks=True)
            main_log_f = os.path.join(args.output, "mutcount_main.log")
            main_log = help_functions.logginginit(args.log_level, main_log_f)
            # initialize mutcount object
            mc = fastq2counts(param_path, updated_out, main_log, args)
            mc._init_skip(skip=True)
    else:
        # alignment
        # make time stamped output folder for this project
        updated_out = os.path.join(args.output, args.name + "_" + time_now)
        os.makedirs(updated_out)  # make directory to save this run
        # write args to file
        args_log_path = os.path.join(updated_out, "args.log")
        write_param(args_log_path, args)

        param_path = os.path.join(updated_out, param_base)
        if not os.path.isfile(param_path):
            param_path = shutil.copy(param_json, updated_out, follow_symlinks=True)

        main_log_f = os.path.join(updated_out, "main.log")
        main_log = help_functions.logginginit(args.log_level, main_log_f)

        # initialize mutcount object
        mc = fastq2counts(param_path, updated_out, main_log, args)
        mc._init_skip(skip=False)

    # start the run
    mc._main()


def write_param(args_log_path, args):
    """
    Write input arguments to param.log
    """
    with open(args_log_path, "w") as args_log:
        for arg in vars(args):
            args_log.write(arg+",")
            args_log.write(f"{getattr(args, arg, 'N/A')}\n")


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='TileSeq mutation counts')
    # user input arguments
    parser.add_argument("-f", "--fastq", help="Path to all fastq files you want to analyze", type=str)
    parser.add_argument("-o", "--output", help="Output folder", type=str, required=True)
    parser.add_argument("-p", "--param", help="csv paramter file", type=str, required=True)
    parser.add_argument("-n", "--name", help="Name for this run", type=str, required=True)
    parser.add_argument("--skip_alignment", action="store_true", help="skip alignment for this analysis, "
                                                                      "ONLY submit jobs for counting mutations in existing output folder")
    parser.add_argument("-r1", help="r1 SAM file", type=str)
    parser.add_argument("-r2", help="r2 SAM file", type=str)

    # user input arguments with default values set
    parser.add_argument("-log", "--log_level", help="set log level: debug, \
        info, warning, error, critical. (default = info)", type=str, default="info")
    parser.add_argument("-env", "--environment", help= "The cluster used to \
        run this script (default = DC)",type=str, default="DC")
    parser.add_argument("-at", type = int, help="Alignment time \
        (default = 8h)", default=8)
    parser.add_argument("-mt", type = int, help="Mutation call time \
        (default = 36h)", default=36)
    parser.add_argument("-mm", type=int, help="Mutation call request memory \
            (default = 15GB)", default=15)
    parser.add_argument("-c", type=int, help="Number of cores", default=16)
    parser.add_argument("-b", "--base", help="ASCII code base", default=33)
    parser.add_argument("-test", action="store_true", help="Turn on testing mode")
    parser.add_argument("-rc", action="store_true", help="Turn on rc mode, both direction of the reads will be "
                                                         "aligned to the reference. Variant calling will be "
                                                         "performed on all the reads that are aligned (BE CAREFUL!)")

    parser.add_argument("-override", "--sr_Override", action="store_true", help="Provide this argument when there is only one replicate")
    parser.add_argument("--posteriorQC", action="store_true", help="Turn on posterior QC mode, this requires more "
                                                                   "memory and runtime, please change the variable "
                                                                   "accordingly")
    parser.add_argument("--wt_override", action="store_true", help="When no wt conditions defined in the parameter sheet, turn on this option will treat EVERYTHING as wt. Phred scores will be adjusted based on the first replicate")

    parser.add_argument("--calibratePhredWT", action="store_true", help="When this parameter is provided, "
                                                                      "use wt to calibrate phred scores")
    parser.add_argument("--calibratePhredPhix", action="store_true", help="When this parameter is provided, "
                                                                      "use phix alignments to calibrate phred scores")
    parser.add_argument("--resubmit", action="store_true", help="For a finished run, batch resubmit failed scripts ("
                                                                "if any)")
    args = parser.parse_args()
    args.environment = args.environment.upper()

    main(args)
